require "bunny"
require "json"
require "watership/version"

module Watership
  CONNECTION_EXCEPTIONS = [
    Bunny::ClientTimeout,
    Bunny::NetworkFailure,
    Bunny::PossibleAuthenticationFailureError,
    Bunny::TCPConnectionFailed
  ]

  CLEAR_CONNECTION_EXCEPTIONS = [
    Bunny::ConnectionError,
    Bunny::ClientTimeout,
    Timeout::Error
  ]

  CLEAR_CHANNEL_EXCEPTIONS = [
    Bunny::ChannelError,
    Timeout::Error
  ]

  def self.register(name, uri, env = 'production')
    registered[name.to_sym] = Client.new(uri, env)
  end

  def self.registered
    @registered_instances ||= {}
  end

  def self.reconnect(name)
    if name == :all
      registered.each do |(name, client)|
        client.reconnect
      end
    else
      registered.fetch(name).reconnect
    end
  end

  def self.close(name)
    if name == :all
      registered.each do |(name, client)|
        client.close
      end
    else
      registered.fetch(name).close
    end
  end

  def self.logger=(logger)
    @logger = logger
  end

  def self.logger
    @logger ||= defined?(Rails) ? Rails.logger : Logger.new(STDOUT)
  end

  def self.enqueue(name, opts)
    registered.fetch(name.to_sym).enqueue(opts)
  end

  class << self
    alias enqueue! enqueue
  end

  class Client
    def initialize(uri, env)
      @uri, @env = uri, env
    end

    def enqueue(options = {})
      if options[:name]
        raise ArgumentError, ":name is no longer valid. Use :queue instead"
      end

      options  = options.dup
      message  = options.delete(:message)
      name     = options.delete(:queue)
      fallback = options.delete(:fallback)

      queue = connect_with_queue(name, options)
      queue.publish(JSON.generate(message))
    rescue *CONNECTION_EXCEPTIONS => exception
      fallback.call if fallback
      notify(exception)
      clear_channel
      logger.error(exception.class.name)
      raise exception
    rescue StandardError => exception
      fallback.call if fallback
      notify(exception)
      logger.error(exception.class.name)
      raise exception
    end

    def connect_with_queue(name, options = {})
      with_channel do |channel|
        channel.queue(name, { durable: true }.merge(options)) if channel
      end
    end

    def thread
      Thread.current[:"watership_#{self.object_id}"] ||= {}
    end

    def close
      clear_channel
      clear_connection
    end

    def reconnect
      close
      create_channel
      true
    end

    def connection
      @bunny_connection ||= begin
        Bunny.new(@uri).tap { |bunny| bunny.start }
      rescue Timeout::Error => exception
        logger.error("Exception #{exception.class.name} got raised in Watership#connection")
        raise exception
      end
    end
    alias create_connection connection

    def clear_connection
      if @bunny_connection.respond_to?(:close)
        begin
          @bunny_connection.close
        rescue *CLEAR_CONNECTION_EXCEPTIONS => exception # the connection is already closed
          logger.error("Exception #{exception.class.name} got raised in Watership#clear_connection")
        end
      end
    ensure
      @bunny_connection = nil
    end

    def with_channel
      channel = create_channel
      yield(channel) if channel
    end

    def create_channel
      thread[:bunny_channel] ||= connection.create_channel
    end

    def clear_channel
      if thread[:bunny_channel].respond_to?(:close)
        begin
          thread[:bunny_channel].close
        rescue *CLEAR_CHANNEL_EXCEPTIONS => exception # the channel is already closed
          logger.error("Exception #{exception.class.name} got raised in Watership#clear_channel")
        end
      end
    ensure
      thread[:bunny_channel] = nil
    end

    def notify(exception)
      Bugsnag.notify(exception) if defined?(Bugsnag) && @env == 'production'
      Airbrake.notify_or_ignore(exception) if defined?(Airbrake) && @env == 'production'
    end

    def logger
      Watership.logger
    end
  end
end
